package net.jaekl.cfb.db.driver;

// Copyright (C) 2015 Christian Jaekl

import static net.jaekl.cfb.db.Column.Null.NOT_NULL;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import net.jaekl.cfb.db.Column;
import net.jaekl.cfb.db.Column.Type;
import net.jaekl.cfb.db.Condition;
import net.jaekl.cfb.db.Row;
import net.jaekl.cfb.db.Sequence;
import net.jaekl.cfb.db.Sort;
import net.jaekl.cfb.db.Table;

public abstract class DbDriver {
	static int PENDING_LIMIT = 1024;	// Rough limit at which point we'll start a new batch for batch updates
	
	DbDriver() {
		
	}
	
	// Load the JDBC driver
	public abstract void load() throws ClassNotFoundException;
	
	public abstract Connection connect(String host, int port, String dbName, String user, String pass) throws SQLException;
	
	public boolean createTable(Connection con, Table table) throws SQLException 
	{
		String sql = createTableSql(table);
		try (PreparedStatement ps = con.prepareStatement(sql)) {
			ps.executeUpdate();
		}
		catch (SQLException exc) {
			throw new SQLException("Failed to executeUpdate:  " + sql, exc);
		}
		
		return true;
	}
	
	public void dropTable(Connection con, Table table) throws SQLException
	{
		String sql = dropTableSql(table);
		try (PreparedStatement ps = con.prepareStatement(sql)) {
			ps.executeUpdate();
		}
		catch (SQLException exc) {
			throw new SQLException("Failed to drop table:  " + sql, exc);
		}
	}
	
	public boolean createSequence(Connection con, Sequence seq) throws SQLException 
	{
		String sql = createSequenceSql(seq);
		try (PreparedStatement ps = con.prepareStatement(sql)) {
			ps.executeUpdate();
		}
		catch (SQLException exc) {
			throw new SQLException("Failed to executeUpdate:  " + sql, exc);
		}
		
		return true;
	}
	
	public void dropSequence(Connection con, Sequence seq) throws SQLException
	{
		String sql = dropSequenceSql(seq);
		try (PreparedStatement ps = con.prepareStatement(sql)) {
			ps.executeUpdate();
		}
		catch (SQLException exc) {
			throw new SQLException("Failed to drop sequence:  " + sql, exc);
		}
	}
	
	public List<Row> select(Connection con, Column[] columns, Table[] tables, Condition[] conditions)
		throws SQLException
	{
		Sort[] sorts = new Sort[0];
		int limit = (-1);	// no limit
		
		return select(con, columns, tables, conditions, sorts, limit);
	}
	
	public List<Row> select(Connection con, Column[] columns, Table[] tables, Condition[] conditions, Sort[] sorts, int limit)
		throws SQLException
	{
		String sql = selectSql(columns, tables, conditions, sorts, limit);
		ArrayList<Row> result = new ArrayList<Row>();
		
		try (PreparedStatement ps = con.prepareStatement(sql)) {
			int index = 0;
			for (Condition condition : conditions) {
				if (condition.getOperation().hasParam()) {
					index++;
					ps.setObject(index, condition.getValue());
				}
			}
			
			try (ResultSet rs = ps.executeQuery()) {
				while (rs.next()) {
					Object[] values = new Object[columns.length];
					for (index = 0; index < columns.length; ++index) {
						if (columns[index].getType().equals(Type.TIMESTAMPTZ)) {
							long milliseconds = rs.getLong(index + 1);
							values[index] = new java.util.Date(milliseconds);
						}
						else {
							values[index] = rs.getObject(index + 1);
						}
					}
					Row row = new Row(columns, values);
					result.add(row);
				}
			}
		}
		
		return result;
	}
	
	// Returns the number of rows inserted
	public int insert(Connection con, Table table, Object[][] values) throws SQLException 
	{
		int count = 0;
		int pendingValues = 0;
		
		String sql = insertSql(table);
		
		try (PreparedStatement ps = con.prepareStatement(sql))
		{
			for (int row = 0; row < values.length; ++row) {
				Object[] data = values[row];
				
				assert(null != data);
				assert(data.length == table.getNumColumns());
				
				for (int col = 0; col < data.length; ++col) {
					Object obj = data[col];
					Column column = table.getColumn(col);
					if (column.getType().equals(Type.TIMESTAMPTZ)) {
						// Special case:  because there's no good way to read a TIMESTAMPTZ from 
						// the database using JDBC, we store it as an integer (milliseconds since
						// the epoch, 01.01.1970 00:00:00.000 UTC).
						Date date = (Date)obj;
						ps.setLong(col + 1, date.getTime());
					}
					else {
						ps.setObject(col + 1, data[col]);
					}
					pendingValues++;
				}
				ps.addBatch();
				
				int rowsFlushed = checkFlushBatch(ps, pendingValues, false);
				if (rowsFlushed > 0) {
					count += rowsFlushed;
					pendingValues = 0;
				}
			}
			
			count += checkFlushBatch(ps, pendingValues, true);
		}
		
		return count;
	}
	
	public long nextVal(Connection con, Sequence seq) throws SQLException
	{
		String sql = nextValSql(seq);
		
		try (PreparedStatement ps = con.prepareStatement(sql)) 
		{
			try (ResultSet rs = ps.executeQuery()) {
				if (rs.next()) {
					return rs.getLong(1);
				}
			}
		}
		
		throw new SQLException("No value returned for sequence:  " + sql);
	}
	
	int checkFlushBatch(PreparedStatement ps, int pendingValues, boolean forceFlush) throws SQLException
 	{
		int count = 0;
		
		if (forceFlush || (pendingValues >= PENDING_LIMIT)) 
		{
			int[] updateCounts = ps.executeBatch();
			for (int i = 0; i < updateCounts.length; ++i) {
				if (updateCounts[i] > 0) {
					count += updateCounts[i];
				}
			}
		}
		
		return count;
	}
	
	String insertSql(Table table) {
		StringBuilder sb = new StringBuilder("INSERT INTO ");
		sb.append(table.getName())
		  .append(" VALUES (");
		
		for (int i = 0; i < table.getNumColumns(); ++i) {
			if (i > 0) {
				sb.append(",");
			}
			sb.append("?");
		}
		sb.append(")");
		
		return sb.toString();
	}
	
	protected String selectSql(Column[] columns, Table[] tables, Condition[] conditions, Sort[] sorts, int limit) 
	{
		StringBuilder sb = new StringBuilder("SELECT ");
		
		boolean firstColumn = true;
		for (Column column : columns) {
			if (firstColumn) {
				firstColumn = false;
			}
			else {
				sb.append(", ");
			}
			sb.append(column.getName());
		}
		
		sb.append(" FROM ");
		
		boolean firstTable = true;
		for (Table table : tables) {
			if (firstTable) {
				firstTable = false;
			}
			else {
				sb.append(", ");
			}
			sb.append(table.getName());
		}
		
		if (null != conditions && conditions.length > 0) {			
			sb.append(" WHERE ");
			
			boolean firstCondition = true;
			
			for (Condition condition : conditions) {
				if (firstCondition) {
					firstCondition = false;
				}
				else {
					sb.append(" AND ");
				}
				
				sb.append(condition.getColumn().getName())
				  .append(condition.getOperation().getSql());
			}
		}

		if (null != sorts && sorts.length > 0) {
			sb.append(" ORDER BY ");
			
			boolean firstSort = true;
			
			for (Sort sort : sorts) {
				if (firstSort) {
					firstSort = false;
				}
				else {
					sb.append(", ");
				}
				
				sb.append(sort.getColumn().getName());
				
				if (sort.getDirection().equals(Sort.Direction.ASCENDING)) {
					sb.append(" ASCENDING ");
				}
				else {
					sb.append(" DESCENDING ");
				}
			}
		}
		
		if (limit > 0) {
			sb.append(" LIMIT " + limit + " ");
		}
		
		return sb.toString();
	}
	
	protected String typeName(Type type) {
		// Special case:  TIMESTAMPTZ stored as INTEGER (milliseconds since the epoch)
		// Reading a TIMESTAMPTZ back from the DB, and converting it to a java.util.Date,
		// is fraught with peril.  The best way around this is to store the dates in 
		// milliseconds-since-the-epoch (01.01.1970 00:00:00.000 UTC).
		if (Type.TIMESTAMPTZ.equals(type)) {
			return Type.INTEGER.toString();
		}
		
		return type.toString();
	}
	
	protected String createColumnSql(Column column) 
	{
		String result = column.getName() + " " + typeName(column.getType());
		if (column.getWidth() > 0) {
			result += "(" + column.getWidth() + ")";
		}
		
		if (NOT_NULL == column.getNull()) {
			result += " NOT NULL";
		}
		else {
			result += " NULL";
		}
	    
		return result;
	}
	
	protected String createTableSql(Table table) 
	{
		assert(null != table);
		assert(null != table.getName());
		assert(table.getNumColumns() > 0);
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("CREATE TABLE ")
		  .append(table.getName())
		  .append("(");
		
		for (int idx = 0; idx < table.getNumColumns(); ++idx) {
			if (idx > 0) {
				sb.append(", ");
			}
			sb.append(createColumnSql(table.getColumn(idx)));
		}
		
		sb.append(")");
		
		return sb.toString();
	}	
	
	protected String dropTableSql(Table table) {
		assert(null != table);
		assert(null != table.getName());
		
		return "DROP TABLE " + table.getName();
	}
	
	protected String createSequenceSql(Sequence seq) {
		assert(null != seq);
		assert(null != seq.getName());
		
		return "CREATE SEQUENCE " + seq.getName();
	}
	
	protected String dropSequenceSql(Sequence seq) {
		assert(null != seq);
		assert(null != seq.getName());
		
		return "DROP SEQUENCE " + seq.getName();
	}
	
	abstract protected String nextValSql(Sequence seq);
}
