package net.jaekl.cfb.xml;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.xml.sax.Attributes;

import net.jaekl.cfb.util.Util;
import net.jaekl.qd.xml.MissingAttributeException;
import net.jaekl.qd.xml.ParseResult;
import net.jaekl.qd.xml.XmlParseException;

public class BugInstance extends ParseResult {

	static final String TAG = "BugInstance";
	static final String[] INTERNAL = {  };
	static final Object[][] EXTERNAL = { { BugClass.TAG, BugClass.class},
		                                 { BugMethod.TAG, BugMethod.class},
		                                 { LocalVariable.TAG, LocalVariable.class},
		                                 { SourceLine.TAG, SourceLine.class} };
	static final String CATEGORY = "category";
	static final String TYPE = "type";

	String m_category;
	String m_type;
	ArrayList<BugClass> m_classes;
	ArrayList<BugMethod> m_methods;
	ArrayList<LocalVariable> m_locals;
	ArrayList<SourceLine> m_lines;
	
	public BugInstance() {
		super(TAG, INTERNAL, EXTERNAL);
		
		m_category = m_type = null;
		m_classes = new ArrayList<BugClass>();
		m_methods = new ArrayList<BugMethod>();
		m_locals = new ArrayList<LocalVariable>();
		m_lines = new ArrayList<SourceLine>();
	}
	
	public String getCategory() { return m_category; }
	public String getType() { return m_type; }
	public List<BugClass> getClasses() { return Collections.unmodifiableList(m_classes); }
	public List<BugMethod> getMethods() { return Collections.unmodifiableList(m_methods); }
	public List<SourceLine> getLines() { return Collections.unmodifiableList(m_lines); }
	public List<LocalVariable> getVariables() { return Collections.unmodifiableList(m_locals); }
	
	@Override
	public void endContents(String uri, String localName, String qName, String chars) 
		throws XmlParseException 
	{
		// no-op
	}

	@Override
	public void endExternal(String uri, String localName, String qName)
		throws XmlParseException 
	{
		if (BugClass.TAG.equals(localName)) {
			ParseResult[] collected = collectParsedChildren(BugClass.class);
			for (ParseResult pr : collected) {
				assert(pr instanceof BugClass);
				m_classes.add((BugClass) pr);
			}
		}
		else if (BugMethod.TAG.equals(localName)) {
			ParseResult[] collected = collectParsedChildren(BugMethod.class);
			for (ParseResult pr : collected) {
				assert(pr instanceof BugMethod);
				m_methods.add((BugMethod)pr);
			}			
		}
		else if (LocalVariable.TAG.equals(localName)) {
			ParseResult[] collected = collectParsedChildren(LocalVariable.class);
			for (ParseResult pr : collected) {
				assert(pr instanceof LocalVariable);
				m_locals.add((LocalVariable)pr);
			}			
		}
		else if (SourceLine.TAG.equals(localName)) {
			ParseResult[] collected = collectParsedChildren(SourceLine.class);
			for (ParseResult pr : collected) {
				assert(pr instanceof SourceLine);
				m_lines.add((SourceLine)pr);
			}
		}
	}
	
	@Override
	public void handleMainAttributes(Attributes attr) throws MissingAttributeException
	{
		m_type = this.getRequiredAttr(TAG, attr, TYPE);
		m_category = this.getRequiredAttr(TAG, attr, CATEGORY);
	}

	@Override 
	public void dump(PrintWriter pw, int indent)
	{
		int childIndent = indent + 2;
		String margin = String.format("%" + indent + "s", "");
		
		pw.println(margin + TAG + " (" + m_type + ")");
		pw.println(margin + CATEGORY + " (" + m_category + ")");
		for (BugClass bc : m_classes) {
			bc.dump(pw, childIndent);
		}
		for (BugMethod bm : m_methods) {
			bm.dump(pw, childIndent);
		}
		for (LocalVariable lv : m_locals) {
			lv.dump(pw, childIndent);
		}
		for (SourceLine sl : m_lines) {
			sl.dump(pw, childIndent);
		}
	}
	
	// Note that this is a heuristic, "fuzzy", equals.
	// Two BugInstances will be considered equal if:
	//   - they refer to the same bug type
	//   - they took place in the same class and method
	//   - the variable names referred to (if any) match
	// In particular, this equality test does not check 
	// for line numbers being equal.  This is by design;
	// we want to consider two bugs to be the "same bug" 
	// even if other changes in the file have shifted 
	// line numbers a bit.
	@Override
	public boolean equals(Object obj) 
	{
		if (null == obj) { 
			return false;
		}
		if (obj instanceof BugInstance) {
			BugInstance that = (BugInstance)obj;
			
			if (! Util.objsAreEqual(this.m_type, that.m_type)) {
				return false;
			}
			
			if (! Util.objsAreEqual(this.m_category, that.m_category)) {
				return false;
			}

			BugMethod thisMethod = this.getPrincipalMethod();
			BugMethod thatMethod = that.getPrincipalMethod();
			if (null == thisMethod) {
				if (null == thatMethod) {
					return false;
				}
			}
			else {
				if (! Util.objsAreEqual(thisMethod.getClassName(), thatMethod.getClassName())) {
					return false;
				}
				if (! Util.objsAreEqual(thisMethod.getMethodName(), thatMethod.getMethodName())) {
					return false;
				}
			}
			
			if (! Util.objsAreEqual(this.getVariables(), that.getVariables())) {
				return false;
			}
			
			return true;
		}
		return false;
	}
	
	@Override
	public int hashCode() 
	{
		int code = Util.objHashCode(m_type)
				 * Util.objHashCode(m_category)
				 * Util.objHashCode(getPrincipalMethod())
				 * Util.objHashCode(getVariables());
		return code;
	}
	
	// Get the "principal" method.
	// This should be the place where the bug is reported.
	BugMethod getPrincipalMethod()
	{
		List<BugMethod> bugMethods = getMethods();
		if ((null == bugMethods) || (0 == bugMethods.size())) {
			return null;
		}
		return bugMethods.get(0);
	}
}
